// -*-Mode: C++;-*- // technically C99

// * BeginRiceCopyright *****************************************************
//
// $HeadURL: http://hpctoolkit.googlecode.com/svn/trunk/src/lib/support-lean/OSUtil.c $
// $Id: OSUtil.c 4059 2013-01-02 20:08:23Z krentel $
//
// --------------------------------------------------------------------------
// Part of HPCToolkit (hpctoolkit.org)
//
// Information about sources of support for research and development of
// HPCToolkit is at 'hpctoolkit.org' and in 'README.Acknowledgments'.
// --------------------------------------------------------------------------
//
// Copyright ((c)) 2002-2013, Rice University
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
// * Redistributions of source code must retain the above copyright
//   notice, this list of conditions and the following disclaimer.
//
// * Redistributions in binary form must reproduce the above copyright
//   notice, this list of conditions and the following disclaimer in the
//   documentation and/or other materials provided with the distribution.
//
// * Neither the name of Rice University (RICE) nor the names of its
//   contributors may be used to endorse or promote products derived from
//   this software without specific prior written permission.
//
// This software is provided by RICE and contributors "as is" and any
// express or implied warranties, including, but not limited to, the
// implied warranties of merchantability and fitness for a particular
// purpose are disclaimed. In no event shall RICE or contributors be
// liable for any direct, indirect, incidental, special, exemplary, or
// consequential damages (including, but not limited to, procurement of
// substitute goods or services; loss of use, data, or profits; or
// business interruption) however caused and on any theory of liability,
// whether in contract, strict liability, or tort (including negligence
// or otherwise) arising in any way out of the use of this software, even
// if advised of the possibility of such damage.
//
// ******************************************************* EndRiceCopyright *

//***************************************************************************
//
// File: 
//   $HeadURL: http://hpctoolkit.googlecode.com/svn/trunk/src/lib/support-lean/OSUtil.c $
//
// Purpose:
//   OS Utilities
//
// Description:
//   [The set of functions, macros, etc. defined in the file]
//
// Author:
//   Nathan Tallent, John Mellor-Crummey, Rice University.
//
//***************************************************************************

//************************* System Include Files ****************************

#include <stdlib.h>

#include <sys/types.h> // getpid()

#define __USE_XOPEN_EXTENDED // for gethostid()
#include <unistd.h>

//*************************** User Include Files ****************************

#include "OSUtil.h"

//*************************** Forward Declarations **************************

//***************************************************************************
// 
//***************************************************************************

uint
OSUtil_pid()
{
  pid_t pid = getpid();
  return (uint)pid;
}


const char*
OSUtil_jobid()
{
  char* jid = NULL;

  // Cobalt
  jid = getenv("COBALT_JOBID");
  if (jid) {
    return jid;
  }

  // PBS
  jid = getenv("PBS_JOBID");
  if (jid) {
    return jid;
  }

  // SLURM
  jid = getenv("SLURM_JOB_ID");
  if (jid) {
    return jid;
  }

  // Sun Grid Engine
  jid = getenv("JOB_ID");
  if (jid) {
    return jid;
  }

  return jid;
}


#define OSUtil_hostid_NULL (-1)

long
OSUtil_hostid()
{
  static long hostid = OSUtil_hostid_NULL;

  if (hostid == OSUtil_hostid_NULL) {
    // gethostid returns a 32-bit id.  treat it as unsigned 
    // to prevent useless sign extension
    hostid = (uint32_t) gethostid();
  }

  return hostid;
}

