// -*-Mode: C++;-*-

// * BeginRiceCopyright *****************************************************
//
// $HeadURL: http://hpctoolkit.googlecode.com/svn/trunk/src/tool/hpcprof-flat/ConfigParser.hpp $
// $Id: ConfigParser.hpp 4059 2013-01-02 20:08:23Z krentel $
//
// --------------------------------------------------------------------------
// Part of HPCToolkit (hpctoolkit.org)
//
// Information about sources of support for research and development of
// HPCToolkit is at 'hpctoolkit.org' and in 'README.Acknowledgments'.
// --------------------------------------------------------------------------
//
// Copyright ((c)) 2002-2013, Rice University
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
// * Redistributions of source code must retain the above copyright
//   notice, this list of conditions and the following disclaimer.
//
// * Redistributions in binary form must reproduce the above copyright
//   notice, this list of conditions and the following disclaimer in the
//   documentation and/or other materials provided with the distribution.
//
// * Neither the name of Rice University (RICE) nor the names of its
//   contributors may be used to endorse or promote products derived from
//   this software without specific prior written permission.
//
// This software is provided by RICE and contributors "as is" and any
// express or implied warranties, including, but not limited to, the
// implied warranties of merchantability and fitness for a particular
// purpose are disclaimed. In no event shall RICE or contributors be
// liable for any direct, indirect, incidental, special, exemplary, or
// consequential damages (including, but not limited to, procurement of
// substitute goods or services; loss of use, data, or profits; or
// business interruption) however caused and on any theory of liability,
// whether in contract, strict liability, or tort (including negligence
// or otherwise) arising in any way out of the use of this software, even
// if advised of the possibility of such damage.
//
// ******************************************************* EndRiceCopyright *

#ifndef ConfigParser_hpp
#define ConfigParser_hpp

//************************ System Include Files ******************************

#include <string>

//*********************** Xerces Include Files *******************************

#include <xercesc/parsers/XercesDOMParser.hpp>
using XERCES_CPP_NAMESPACE::XercesDOMParser;

#include <xercesc/dom/DOMNode.hpp> 
using XERCES_CPP_NAMESPACE::DOMNode;

//************************* User Include Files *******************************

#include <lib/analysis/Args.hpp>

#include <lib/prof/Metric-Mgr.hpp>

#include <lib/profxml/XercesErrorHandler.hpp>

#include <lib/support/diagnostics.h>

//************************ Forward Declarations ******************************

//****************************************************************************
// 
//***************************************************************************

class ConfigParser {
public:
  ConfigParser(const std::string& inputFile, XercesErrorHandler &errHndlr);
  ~ConfigParser();

  void parse(Analysis::Args& args, Prof::Metric::Mgr& metricMgr);

private:
  XercesDOMParser* m_parser;
  DOMNode* m_doc;
};


//***************************************************************************

#define ConfigParser_Throw(streamArgs) DIAG_ThrowX(ConfigParserException, streamArgs)

class ConfigParserException : public Diagnostics::Exception {
public:
  ConfigParserException(const std::string x,
		      const char* filenm = NULL, unsigned int lineno = 0)
    : Diagnostics::Exception(x, filenm, lineno)
    { }
  
  virtual std::string message() const { 
    return "CONFIGURATION file error [ConfigParserException]: " + what();
  }

private:
};


#endif /* ConfigParser_hpp */
